# This file is part of Moksha.
# Copyright (C) 2008-2014  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
:mod:`moksha.hub.api.producer - The Moksha Producer API
=======================================================
"""

import logging
import time

from datetime import timedelta

import moksha.hub.reactor
from moksha.common.lib.helpers import create_app_engine

log = logging.getLogger('moksha.hub')


class Producer(object):
    """ The parent Producer class. """

    # Internal use only
    _initialized = False
    _exception_count = 0

    def __init__(self, hub):
        self.hub = hub
        self.log = log

        # If the stream specifies an 'app', then setup `self.engine` to
        # be a SQLAlchemy engine for that app, along with a configured
        # DBSession
        app = getattr(self, 'app', None)
        self.engine = self.DBSession = None
        if app:
            log.debug("Setting up individual engine for producer")
            from sqlalchemy.orm import sessionmaker
            self.engine = create_app_engine(app)
            self.DBSession = sessionmaker(bind=self.engine)()

        self._initialized = True

    def __json__(self):
        return {
            "name": type(self).__name__,
            "module": type(self).__module__,
            "initialized": self._initialized,
            "exceptions": self._exception_count,
        }

    def send_message(self, topic, message):
        try:
            self.hub.send_message(topic, message)
        except Exception as e:
            log.error('Cannot send message: %s' % e)

    def stop(self):
        if hasattr(self, 'hub') and self.hub:
            self.hub.close()
        if hasattr(self, 'DBSession') and self.DBSession:
            self.DBSession.close()


class PollingProducer(Producer):
    """ A self-polling producer

    This class represents a data stream that wakes up at a given frequency,
    and calls the :meth:`poll` method.
    """
    frequency = None  # Either a timedelta object, or the number of seconds
    now = False
    die = False

    def __init__(self, hub):
        super(PollingProducer, self).__init__(hub)

        if isinstance(self.frequency, timedelta):
            self.frequency = self.frequency.seconds + \
                (self.frequency.days * 24 * 60 * 60) + \
                (self.frequency.microseconds / 1000000.0)

        self._last_ran = None
        # This is used to determine if the configured frequency has been meet,
        # and the poller should run. This allows the poller to check if
        # self.die is True more frequently.
        self._until_next_poll = self.frequency

        log.debug("Setting a %s second timer" % self.frequency)
        moksha.hub.reactor.reactor.callInThread(self._work)

    def __json__(self):
        data = super(PollingProducer, self).__json__()
        data.update({
            "frequency": self.frequency,
            "now": self.now,
            "last_ran": self._last_ran,
        })
        return data

    def poll(self):
        raise NotImplementedError

    def _poll(self):
        self._last_ran = time.time()
        try:
            self.poll()
            self._exception_count = 0  # Reset to 0 if things are gravy
        except Exception:
            # Otherwise, keep track of how many exceptions we hit in a row
            self._exception_count = self._exception_count + 1
            # And re-raise the exception so it can be logged.
            raise

    def _work(self):
        # If asked to, we can fire immediately at startup
        if self.now:
            self._poll()

        while not self.die:
            if not self.frequency:
                # If no frequency is set, just continuously poll
                self._poll()
                continue

            # If _until_next_poll is less than or equal to 0, that means
            # the frequency has been met and the poller needs to run again
            if self._until_next_poll <= 0:
                self._poll()
                # Reset _until_next_poll so that polling doesn't happen
                # until the frequency is met again
                self._until_next_poll = self.frequency
            else:
                # Only sleep 5 seconds (or _until_next_poll if it's shorter) at
                # a time since we want to check if the poller should die
                # frequently. Otherwise, you'll end up with Moksha Hub in a
                # state where the hub is stopped but the poller is still
                # sleeping until the frequency is met
                sleep_time = min((5, self._until_next_poll))
                time.sleep(sleep_time)
                self._until_next_poll -= sleep_time

    def stop(self):
        super(PollingProducer, self).stop()
        self.die = True



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test Moksha's Hub """

import threading
import moksha

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from time import sleep, time
from uuid import uuid4
from kitchen.iterutils import iterate
import tempfile
import shutil
import os
import stat
import zmq
import json

import moksha.common.testtools.utils as testutils

import moksha.hub.api
from moksha.hub.hub import MokshaHub, CentralMokshaHub
from moksha.hub.reactor import reactor as _reactor
from moksha.hub.monitoring import MonitoringProducer
from nose.tools import (eq_, assert_true, assert_false)


# Some constants used throughout the hub tests
sleep_duration = 0.25
secret = "secret_message"


def simulate_reactor(duration=sleep_duration):
    """ Simulate running the reactor for `duration` milliseconds """
    global _reactor
    start = time()
    while time() - start < duration:
        _reactor.doPoll(0.0001)
        _reactor.runUntilCurrent()


class TestHub:

    def _setUp(self):
        def kernel(config):
            self.hub = MokshaHub(config=config)
            self.topic = str(uuid4())

        for __setup, name in testutils.make_setup_functions(kernel):
            yield __setup, name

    def _tearDown(self):
        self.hub.close()

    @testutils.crosstest
    def test_hub_creation(self):
        """ Test that we can simply create the hub. """
        assert_true(self.hub)
        eq_(self.hub.topics, {})

    @testutils.crosstest
    def test_hub_send_recv(self):
        "Test that we can send a message and receive it."

        messages_received = []

        def callback(json):
            messages_received.append(json.body[1:-1])

        self.hub.subscribe(topic=self.topic, callback=callback)
        sleep(sleep_duration)

        self.hub.send_message(topic=self.topic, message=secret)

        simulate_reactor(sleep_duration)
        sleep(sleep_duration)

        eq_(messages_received, [secret])

    @testutils.crosstest
    def test_hub_no_subscription(self):
        "Test that we don't receive messages we're not subscribed for."

        messages_received = []

        def callback(json):
            messages_received.append(json.body[1:-1])

        self.hub.send_message(topic=self.topic, message=secret)
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        eq_(messages_received, [])


class TestConsumer:

    def _setUp(self):
        def kernel(config):
            self.hub = MokshaHub(config=config)
            self.a_topic = str(uuid4())

        for __setup, name in testutils.make_setup_functions(kernel):
            yield __setup, name

    def _tearDown(self):
        self.hub.close()

    def fake_register_consumer(self, cons):
        """ Fake register a consumer, not by entry-point like usual.

        Normally, consumers are identified by the hub by way of entry-points
        Ideally, this test would register the TestConsumer on the
        moksha.consumers entry point, and the hub would pick it up.
        I'm not sure how to do that, so we're going to fake it and manually
        add this consumer to the list of consumers of which the Hub is aware.
        """
        consumer = cons(self.hub)
        consume = consumer.consume
        for topic in iterate(cons.topic):
            self.hub.topics[topic] = self.hub.topics.get(topic, [])
            if consume not in self.hub.topics[topic]:
                print('registering fake topic %r' % topic)
                self.hub.topics[topic].append(consume)
        sleep(sleep_duration)
        return consumer

    @testutils.crosstest
    def test_abstract(self):
        """ Ensure that conumsers with no consume method raise exceptions. """

        class StillAbstractConsumer(moksha.hub.api.consumer.Consumer):
            pass

        try:
            c = StillAbstractConsumer(self.hub)
            c.consume("foo")
            assert(False)
        except NotImplementedError as e:
            pass

    @testutils.crosstest
    def test_receive_without_json(self):
        """ Try sending/receiving messages without jsonifying. """

        messages_received = []

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            jsonify = False
            topic = self.a_topic

            def _consume(self, message):
                messages_received.append(message)

        self.fake_register_consumer(TestConsumer)

        # Now, send a generic message to that topic, and see if we get one.
        self.hub.send_message(topic=self.a_topic, message=secret)
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        eq_(len(messages_received), 1)

    @testutils.crosstest
    def test_receive_str(self):
        """ Send a message  Consume and verify it. """

        messages_received = []

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def _consume(self, message):
                messages_received.append(message['body'])

        self.fake_register_consumer(TestConsumer)

        # Now, send a generic message to that topic, and see if the consumer
        # processed it.
        self.hub.send_message(topic=self.a_topic, message=secret)
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        eq_(messages_received, [secret])

    @testutils.crosstest
    def test_receive_str_double(self):
        """ Send a message.  Have two consumers consume it. """

        messages_received = []

        class TestConsumer1(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def _consume(self, message):
                messages_received.append(message['body'])

        class TestConsumer2(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def _consume(self, message):
                messages_received.append(message['body'])

        self.fake_register_consumer(TestConsumer1)
        self.fake_register_consumer(TestConsumer2)

        # Now, send a generic message to that topic, and see if the consumer
        # processed it.
        self.hub.send_message(topic=self.a_topic, message=secret)
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        eq_(messages_received, [secret, secret])

    @testutils.crosstest
    def test_receive_str_near_miss(self):
        """ Send a message.  Three consumers.  Only one receives. """

        messages_received = []

        class BaseConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def _consume(self, message):
                messages_received.append(message['body'])

        class Consumer1(BaseConsumer):
            pass

        class Consumer2(BaseConsumer):
            topic = BaseConsumer.topic[:-1]

        class Consumer3(BaseConsumer):
            topic = BaseConsumer.topic + "X"

        self.fake_register_consumer(Consumer1)
        self.fake_register_consumer(Consumer2)
        self.fake_register_consumer(Consumer3)

        # Now, send a generic message to that topic, and see if Consumer1
        # processed it but that Consumer2 and Consumer3 didn't
        self.hub.send_message(topic=self.a_topic, message=secret)
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        eq_(messages_received, [secret])

    @testutils.crosstest
    def test_receive_dict(self):
        """ Send a dict with a message.  Consume, extract, and verify it. """

        obj = {'secret': secret}
        messages_received = []

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def _consume(self, message):
                obj = message['body']
                messages_received.append(obj['secret'])

        self.fake_register_consumer(TestConsumer)

        # Now, send a generic message to that topic, and see if the consumer
        # processed it.
        self.hub.send_message(topic=self.a_topic, message=obj)
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        eq_(messages_received, [secret])

    @testutils.crosstest
    def test_receive_n_messages(self):
        """ Send `n` messages, receive `n` messages. """

        n_messages = 10
        messages_received = []

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def _consume(self, message):
                messages_received.append(message['body'])

        self.fake_register_consumer(TestConsumer)

        # Now, send n messages and make sure that n messages were consumed.
        for i in range(n_messages):
            self.hub.send_message(topic=self.a_topic, message=secret)

        simulate_reactor(sleep_duration)
        sleep(sleep_duration)

        eq_(len(messages_received), n_messages)

    @testutils.crosstest
    def test_receive_n_dicts(self):
        """ Send `n` dicts, receive `n` dicts. """

        n_messages = 10
        obj = {'secret': secret}
        messages_received = []

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def _consume(self, message):
                messages_received.append(message['body'])

        self.fake_register_consumer(TestConsumer)

        # Now, send n objects and make sure that n objects were consumed.
        for i in range(n_messages):
            self.hub.send_message(topic=self.a_topic, message=obj)

        simulate_reactor(sleep_duration)
        sleep(sleep_duration)

        eq_(len(messages_received), n_messages)

    @testutils.crosstest
    def test_multiple_topics(self):
        """ Send a message to multiple topics. """
        n_messages = 2
        obj = {'secret': secret}
        messages_received = []
        b_topic = str(uuid4())

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = [self.a_topic, b_topic]

            def _consume(self, message):
                messages_received.append(message['body'])

        self.fake_register_consumer(TestConsumer)

        self.hub.send_message(topic=self.a_topic, message=obj)
        self.hub.send_message(topic=b_topic, message=obj)

        simulate_reactor(sleep_duration)
        sleep(sleep_duration)

        eq_(len(messages_received), n_messages)

    @testutils.crosstest
    def test_dynamic_topic(self):
        """ Test that a topic can be set at runtime (not import time) """

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = "bad topic"

            def __init__(self, *args, **kw):
                super(TestConsumer, self).__init__(*args, **kw)
                self.topic = "good topic"

            def _consume(self, message):
                pass

        # Just a little fake config.
        config = dict(
            zmq_enabled=True,
            zmq_subscribe_endpoints='',
            zmq_published_endpoints='',
        )
        central = CentralMokshaHub(config, [TestConsumer], [])

        # Guarantee that "bad topic" is not in the topics list.
        eq_(list(central.topics.keys()), ["good topic"])

    @testutils.crosstest
    def test_open_and_close(self):
        """ Test that a central hub with a consumer can be closed.. ;) """

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = "whatever"

            def _consume(self, message):
                pass

        # Just a little fake config.
        config = dict(
            zmq_enabled=True,
            zmq_subscribe_endpoints='',
            zmq_published_endpoints='',
        )
        central = CentralMokshaHub(config, [TestConsumer], [])
        central.close()

    @testutils.crosstest
    def test_consumer_stats_queued(self):
        """ Verify that message processing stats are set for queued messages. """

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def consume(self, message):
                pass

        cons = self.fake_register_consumer(TestConsumer)

        for i in range(5):
            self.hub.send_message(topic=self.a_topic, message=secret)

        simulate_reactor(sleep_duration)
        sleep(sleep_duration)

        eq_(cons.headcount_in, 5)
        eq_(cons.headcount_out, 0)
        eq_(cons._exception_count, 0)
        eq_(len(cons._times), 0)

    @testutils.crosstest
    def test_consumer_stats_processed(self):
        """ Verify that message processing stats are set for processed messages. """

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def consume(self, message):
                pass

        self.hub.config['moksha.blocking_mode'] = True
        cons = self.fake_register_consumer(TestConsumer)

        for i in range(5):
            self.hub.send_message(topic=self.a_topic, message=secret)

        simulate_reactor(sleep_duration)
        sleep(sleep_duration)

        eq_(cons.headcount_in, 5)
        eq_(cons.headcount_out, 5)
        eq_(cons._exception_count, 0)
        eq_(len(cons._times), 5)

    @testutils.crosstest
    def test_consumer_stats_exceptions(self):
        """ Verify that message processing stats are set for messages that generate exceptions. """

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def consume(self, message):
                if message['body'] % 2:
                    raise RuntimeError()

        self.hub.config['moksha.blocking_mode'] = True
        cons = self.fake_register_consumer(TestConsumer)

        for i in range(5):
            self.hub.send_message(topic=self.a_topic, message=i)

        simulate_reactor(sleep_duration)
        sleep(sleep_duration)

        eq_(cons.headcount_in, 5)
        eq_(cons.headcount_out, 5)
        eq_(cons._exception_count, 2)
        eq_(len(cons._times), 5)

    @testutils.crosstest
    def test_consumer_stats_overflow(self):
        """ Verify that Consumer._times doesn't grow beyond a maximum size. """

        class TestConsumer(moksha.hub.api.consumer.Consumer):
            topic = self.a_topic

            def consume(self, message):
                pass

        self.hub.config['moksha.blocking_mode'] = True
        cons = self.fake_register_consumer(TestConsumer)

        for i in range(1500):
            self.hub.send_message(topic=self.a_topic, message=secret)

        simulate_reactor(sleep_duration)
        sleep(sleep_duration)

        eq_(cons.headcount_in, 1500)
        eq_(cons.headcount_out, 1500)
        eq_(cons._exception_count, 0)
        eq_(len(cons._times), 1024)


class TestProducer:
    def _setUp(self):
        def kernel(config):
            self.hub = MokshaHub(config=config)
            self.a_topic = a_topic = str(uuid4())

        for __setup, name in testutils.make_setup_functions(kernel):
            yield __setup, name

    def _tearDown(self):
        self.hub.close()

    def fake_register_producer(self, prod):
        """ Fake register a producer, not by entry-point like usual.

        Registering producers is a little easier than registering consumers.
        The MokshaHub doesn't even keep track of the .poll method callbacks.
        We simply instantiate the producer (and it registers itself with the
        hub).
        """
        return prod(self.hub)

    @testutils.crosstest
    def test_produce_ten_strs(self):
        """ Produce ten-ish strings. """

        messages_received = []

        class TestProducer(moksha.hub.api.producer.PollingProducer):
            topic = self.a_topic
            frequency = sleep_duration / 10.9
            called = 0

            def poll(self):
                self.called = self.called + 1

        # Ready?
        prod = self.fake_register_producer(TestProducer)

        def killer():
            sleep(sleep_duration)
            prod.die = True

        threading.Thread(target=killer).start()
        prod._work()

        # Finally, the check.  Did we get our ten messages? (or about as much)
        assert prod.called > 8
        assert prod.called < 12

    @testutils.crosstest
    def test_idempotence(self):
        """ Test that running the same test twice still works. """
        return self.test_produce_ten_strs()


class TestMonitoring:
    def _setUp(self):
        def kernel(config):
            self.hub = CentralMokshaHub(config=config)
            self.a_topic = a_topic = str(uuid4())

        for __setup, name in testutils.make_setup_functions(kernel):
            yield __setup, name

    def _tearDown(self):
        self.hub.close()

    @testutils.crosstest
    def test_monitoring(self):
        """ Test that the MonitoringProducer works as expected. """
        tmpdir = tempfile.mkdtemp()
        try:
            zmq_file = tmpdir + '/socket'
            zmq_socket = 'ipc://' + zmq_file
            self.hub.config['moksha.monitoring.socket'] = zmq_socket
            self.hub.config['moksha.monitoring.socket.mode'] = '777'
            mon = MonitoringProducer(self.hub)
            assert_true(os.path.exists(zmq_file))
            assert_true(stat.S_IMODE(os.stat(zmq_file).st_mode) == 0o777)
            ctx = zmq.Context()
            sub = ctx.socket(zmq.SUB)
            sub.setsockopt(zmq.RCVTIMEO, 10000)
            sub.setsockopt_string(zmq.SUBSCRIBE, u'')
            sub.connect(zmq_socket)
            data = []
            def recv():
                data.append(sub.recv())
            thread = threading.Thread(target=recv)
            thread.start()
            sleep(sleep_duration)
            mon.poll()
            thread.join()
            eq_(len(data), 1)
            d = json.loads(data[0])
            eq_(len(d['consumers']), 0)
            eq_(len(d['producers']), 1)
            eq_(d['producers'][0]['name'], 'MonitoringProducer')
        finally:
            shutil.rmtree(tmpdir)



# This file is part of Moksha.
# Copyright (C) 2008-2013  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Test Moksha's Websocket Server.

ZeroMQ only for now.
"""

import moksha
import json
import websocket
import copy

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from time import sleep, time
from uuid import uuid4

from moksha.hub.hub import CentralMokshaHub
from moksha.hub.reactor import reactor as _reactor
from nose.tools import eq_, assert_true, assert_false, raises


# TODO -- these are duplicated in test_hub.. they should be imported
# Some constants used throughout the hub tests
sleep_duration = 0.25
secret = "secret_message"


# TODO -- this is duplicated in test_hub.  It should be imported from a common
# place.
def simulate_reactor(duration=sleep_duration):
    """ Simulate running the reactor for `duration` milliseconds """
    global _reactor
    start = time()
    while time() - start < duration:
        _reactor.doPoll(0.0001)
        _reactor.runUntilCurrent()


class TestWebSocketServer(unittest.TestCase):

    def setUp(self):
        config = {
            'moksha.livesocket': True,
            'moksha.livesocket.backend': 'websocket',
            'moksha.socket.notify': True,
            'moksha.livesocket.websocket.port': 8009,
            "zmq_publish_endpoints": "tcp://*:6543",
            "zmq_subscribe_endpoints": "tcp://127.0.0.1:6543",
            "zmq_enabled": True,
            'zmq_strict': False,
        }
        self.hub = CentralMokshaHub(config=config)
        self.topic = str(uuid4())

    def tearDown(self):
        self.hub.close()

        if hasattr(self.hub, 'websocket_server'):
            retval = self.hub.websocket_server.stopListening()

        # It can take some time to unregister our WS server from its port
        simulate_reactor(sleep_duration)

    def test_ws_subscribe_and_recv(self):
        """ Test that we can subscribe for and receive a message. """

        self.received_message = None
        import threading

        class client_thread(threading.Thread):
            def run(thread):
                ws = websocket.WebSocket()
                ws.settimeout(5)
                ws.connect("ws://127.0.0.1:{port}/".format(
                    port=self.hub.config['moksha.livesocket.websocket.port'],
                ))

                ws.send(json.dumps(dict(
                    topic="__topic_subscribe__",
                    body=self.topic,
                )))

                # Receive that..
                message = ws.recv()
                self.received_message = json.loads(message)['body']
                ws.close()

        client = client_thread()
        client.start()

        # Process the connection from the client-thread.
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        simulate_reactor(sleep_duration)

        # Now, send a message...
        self.hub.send_message(
            topic=self.topic,
            message=secret,
        )

        # Process the sending of our special message.
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        simulate_reactor(sleep_duration)

        client.join()
        eq_(self.received_message, secret)

    def test_ws_subscribe_multiple(self):
        """ Test that we can subscribe to a few different topics. """

        self.received_messages = []
        import threading

        num_topics = 3

        class client_thread(threading.Thread):
            def run(thread):
                ws = websocket.WebSocket()
                ws.settimeout(5)
                ws.connect("ws://127.0.0.1:{port}/".format(
                    port=self.hub.config['moksha.livesocket.websocket.port'],
                ))

                for i in range(num_topics):
                    ws.send(json.dumps(dict(
                        topic="__topic_subscribe__",
                        body=self.topic + "_" + str(i),
                    )))

                # Receive that..
                for i in range(num_topics):
                    try:
                        self.received_messages.append(
                            json.loads(ws.recv())['body']
                        )
                    except Exception:
                        pass

                ws.close()

        client = client_thread()
        client.start()

        # Process the connection from the client-thread.
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        simulate_reactor(sleep_duration)

        # Now, send a message...
        for i in range(num_topics):
            self.hub.send_message(
                topic=self.topic + "_" + str(i),
                message=secret,
            )

        # Process the sending of our special message.
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        simulate_reactor(sleep_duration)

        client.join()
        eq_(self.received_messages, [secret] * num_topics)

    def test_ws_subscribe_filter(self):
        """ Test that the WS server only sends desired topics. """

        self.received_messages = []
        import threading

        num_topics = 1

        class client_thread(threading.Thread):
            def run(thread):
                ws = websocket.WebSocket()
                ws.settimeout(5)
                ws.connect("ws://127.0.0.1:{port}/".format(
                    port=self.hub.config['moksha.livesocket.websocket.port'],
                ))

                for i in range(num_topics):
                    ws.send(json.dumps(dict(
                        topic="__topic_subscribe__",
                        body=self.topic + "_" + str(i),
                    )))

                # Receive that..
                for i in range(num_topics + 1):
                    try:
                        self.received_messages.append(
                            json.loads(ws.recv())['body']
                        )
                    except Exception:
                        pass

                ws.close()

        client = client_thread()
        client.start()

        # Process the connection from the client-thread.
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        simulate_reactor(sleep_duration)

        # Now, send a message...
        for i in range(num_topics + 1):
            self.hub.send_message(
                topic=self.topic + "_" + str(i),
                message=secret,
            )

        # Process the sending of our special message.
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        simulate_reactor(sleep_duration)

        client.join()
        eq_(self.received_messages, [secret] * num_topics)

    def test_ws_multiple_clients_different_topics(self):
        """ Test that the WS server can differentiate clients. """

        import threading

        num_topics = 2

        class client_thread(threading.Thread):
            def run(thread):
                thread.received_messages = []
                ws = websocket.WebSocket()
                ws.settimeout(5)
                ws.connect("ws://127.0.0.1:{port}/".format(
                    port=self.hub.config['moksha.livesocket.websocket.port'],
                ))

                for i in range(num_topics):
                    ws.send(json.dumps(dict(
                        topic="__topic_subscribe__",
                        body=thread.topic + "_" + str(i),
                    )))

                # Receive that..
                for i in range(num_topics + 2):
                    try:
                        thread.received_messages.append(
                            json.loads(ws.recv())['body']
                        )
                    except Exception:
                        pass

                ws.close()

        client1 = client_thread()
        client2 = client_thread()
        client1.topic = self.topic + "_topic_1"
        client2.topic = self.topic + "_topic_2"
        client1.received_messages = []
        client2.received_messages = []
        client1.start()
        client2.start()

        # Process the connection from the client-thread.
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        simulate_reactor(sleep_duration)

        # Now, send a message...
        for i in range(num_topics):
            self.hub.send_message(
                topic=self.topic + "_topic_1" + "_" + str(i),
                message=secret + "_1",
            )
            self.hub.send_message(
                topic=self.topic + "_topic_2" + "_" + str(i),
                message=secret + "_2",
            )

        # Process the sending of our special message.
        simulate_reactor(sleep_duration)
        sleep(sleep_duration)
        simulate_reactor(sleep_duration)

        client1.join()
        client2.join()
        eq_(client1.received_messages, [secret + "_1"] * num_topics)
        eq_(client2.received_messages, [secret + "_2"] * num_topics)


if __name__ == '__main__':
    unittest.main()



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Luke Macken <lmacken@redhat.com>

import logging
from moksha.common.lib.converters import asbool


from moksha.hub.amqp.base import BaseAMQPHubExtension

log = logging.getLogger(__name__)

NONPERSISTENT_DELIVERY = PERSISTENT_DELIVERY = range(1, 3)


class AMQPLibHubExtension(BaseAMQPHubExtension):
    """ An AMQPHub implemention using the amqplib module """

    def __init__(self, hub, config):
        import amqplib.client_0_8 as amqp

        self.config = config

        broker = self.config.get('amqp_broker')
        ssl = asbool(self.config.get('amqp_broker_ssl', False))
        use_threading = asbool(self.config.get('amqp_broker_threaded', False))
        username = self.config.get('amqp_broker_username', 'guest')
        password = self.config.get('amqp_broker_password', 'guest')

        self.conn = amqp.Connection(
            host=broker,
            ssl=ssl,
            use_threading=use_threading,
            userid=username,
            password=password
        )
        self.channel = self.conn.channel()
        self.channel.access_request(
            '/data', active=True, write=True, read=True)
        super(AMQPLibHubExtension, self).__init__()

    def create_queue(self, queue, exchange='amq.fanout', durable=True,
                     exclusive=False, auto_delete=False):
        """ Declare a `queue` and bind it to an `exchange` """
        if not queue in self.queues:
            log.info("Creating %s queue" % queue)
            self.channel.queue_declare(queue,
                                       durable=durable,
                                       exclusive=exclusive,
                                       auto_delete=auto_delete)

    def exchange_declare(self, exchange, type='fanout', durable=True,
                         auto_delete=False):
        self.channel.exchange_declare(exchange=exchange, type=type,
                                      durable=durable, auto_delete=auto_delete)

    def queue_bind(self, queue, exchange, routing_key=''):
        self.channel.queue_bind(queue, exchange, routing_key=routing_key)

    def send_message(self, topic, message, **headers):
        """
        Send an AMQP message to a given exchange with the specified routing key
        """
        import amqplib.client_0_8 as amqp

        msg = amqp.Message(message, **headers)
        msg.properties["delivery_mode"] = headers.get(
            "delivery_mode", PERSISTENT_DELIVERY)
        self.channel.basic_publish(
            msg,
            headers.get('exchange', 'amq.topic'),
            routing_key=topic
        )
        super(AMQPLibHubExtension, self).send_message(
            topic, message, **headers)

    def subscribe(self, topic, callback):
        queue_name = str(uuid.uuid4())
        self.queue_declare(queue=queue_name, exclusive=True,
                           auto_delete=True)
        self.exchange_bind(queue_name, binding_key=topic)
        self.queue_subscribe(queue_name, callback)
        super(AMQPLibHubExtension, self).subscribe(topic, callback)

    def get_message(self, queue):
        """ Immediately grab a message from the queue.

        This call will not block, and will return None if there are no new
        messages in the queue.
        """
        msg = self.channel.basic_get(queue, no_ack=True)
        return msg

    def queue_subscribe(self, queue, callback, no_ack=True):
        """
        Consume messages from a given `queue`, passing each to `callback`
        """
        self.channel.basic_consume(queue, callback=callback, no_ack=no_ack)

    def wait(self):
        self.channel.wait()

    def close(self):
        try:
            if hasattr(self, 'channel') and self.channel:
                self.channel.close()
        except Exception as e:
            log.exception(e)
        try:
            if hasattr(self, 'conn') and self.conn:
                self.conn.close()
        except Exception as e:
            log.exception(e)



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Luke Macken <lmacken@redhat.com>
#
# Based on code from stomper's examples
# (c) Oisin Mulvihill, 2007-07-26.
# License: http://www.apache.org/licenses/LICENSE-2.0

import logging

from distutils.version import LooseVersion

from moksha.common.lib.converters import asbool
from moksha.hub.reactor import reactor

try:
    # stomper is not ready for py3
    try:
        # Try first to use modern stomp-1.1
        import stomper.stomp_11 as stomper
    except ImportError:
        # Failing that, use whatever is available.
        try:
            import stomper
        except ImportError:
            pass
    from stomper.stompbuffer import StompBuffer
    from twisted.internet.protocol import Protocol
    class Base(Protocol, stomper.Engine):
        pass
except ImportError:
    Base = object

log = logging.getLogger(__name__)

class StompProtocol(Base):

    def __init__(self, client, username='', password=''):
        stomper.Engine.__init__(self)
        self.username = username
        self.password = password
        self.counter = 1
        self.client = client
        self.buffer = StompBuffer()

    def connected(self, msg):
        """Once connected, subscribe to message queues """
        stomper.Engine.connected(self, msg)
        log.info("StompProtocol Connected: session %s." %
                 msg['headers']['session'])

        # https://stomp.github.io/stomp-specification-1.1.html#Heart-beating
        server_heartbeat = msg['headers'].get('heart-beat', 0)
        if server_heartbeat:
            log.debug("(server wants heart-beat (%s))" % server_heartbeat)
            sx, sy = server_heartbeat.split(',')
            server_heartbeat = int(sy)

        self.client.connected(server_heartbeat)

    def subscribe(self, dest, **headers):
        f = stomper.Frame()
        # https://stomp.github.io/stomp-specification-1.2.html#SUBSCRIBE_ack_Header
        ack = self.client.hub.config.get('stomp_ack_mode', 'auto')
        if stomper.STOMP_VERSION != '1.0':
            f.unpack(stomper.subscribe(dest, dest, ack=ack))
        else:
            f.unpack(stomper.subscribe(dest, ack=ack))
        f.headers.update(headers)
        cmd = f.pack()
        log.debug(cmd)
        self.transport.write(cmd.encode('utf-8'))

    def connectionMade(self):
        """ Register with stomp server """
        log.debug("Connecting with stomp-%s" % stomper.STOMP_VERSION)
        if stomper.STOMP_VERSION != '1.0':
            host, port = self.client.addresses[self.client.address_index]
            interval = (self.client.client_heartbeat, 0)
            log.debug("(proposing heartbeat of (%i,%i))" % interval)
            cmd = stomper.connect(self.username, self.password, host, interval)
        else:
            cmd = stomper.connect(self.username, self.password)
        log.debug(cmd)
        self.transport.write(cmd.encode('utf-8'))

    def error(self, msg):
        """ Extend stomper's own error method to kill the hub. """
        super(StompProtocol, self).error(msg)
        log.error("Requesting shutdown of hub for STOMP error.")
        reactor.callLater(0, self.client.hub.close)
        reactor.callLater(0, reactor.stop)

    def ack(self, msg):
        """ Override stomper's own ack to be smarter, based on mode. """
        # stomper does the incorrect thing when the ack mode is auto.  It acks
        # every message, regardless of the mode.  However, if the mode is
        # 'auto', then we should *not* send acks.  Here, make sure we don't
        # send an ack in that mode.
        if self.client.hub.config.get('stomp_ack_mode', 'auto') == 'auto':
            return stomper.NO_REPONSE_NEEDED

        # Otherwise, do what stomper do if the mode is *not* auto.
        return super(StompProtocol, self).ack(msg)

    def dataReceived(self, data):
        """Data received, react to it and respond if needed """
        self.buffer.appendData(data.decode('utf-8', errors='replace'))
        while True:
           msg = self.buffer.getOneMessage()
           if msg is None:
               break

           handled = self.client.hub.consume_stomp_message(msg)

           # See if stomper thinks we need to send anything back.
           response = self.react(msg)

           # If this kind of message doesn't need any response, then quit.
           if not response:
               log.debug("StompProtocol sending no response to broker.")
               return

           # Otherwise, see if we need to turn a naive 'ack' from stomper into
           # a 'nack' if our consumers failed to do their jobs.
           if handled is False and response.startswith("ACK\n"):

               send_nacks = asbool(self.client.hub.config.get('stomp_send_explicit_nacks', True))
               if not send_nacks:
                   log.warn("Message handling failed.  stomp_send_explicit_nacks=%r.  "
                            "Sending no reply to the broker.", send_nacks)
                   # Return, so as not to send an erroneous ack.
                   return

               if LooseVersion(stomper.STOMP_VERSION) < LooseVersion('1.1'):
                   log.error("Unable to NACK stomp %r" % stomper.STOMP_VERSION)
                   # Also, not sending an erroneous ack.
                   return

               message_id = msg['headers']['message-id']
               subscription = msg['headers']['subscription']
               transaction_id = msg['headers'].get('transaction-id')
               response = stomper.nack(message_id, subscription, transaction_id)

           # Finally, send our response (ACK or NACK) back to the broker.
           if not handled:
               log.warn("handled=%r.  Responding with %s" % (handled, response))
           else:
               log.debug("handled=%r.  Responding with %s" % (handled, response))
           self.transport.write(response.encode('utf-8'))
